/*  zncurses, a zsh module for using ncurses windows
    Copyright (C) 2007  Clint Adams

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#define _XOPEN_SOURCE_EXTENDED 1

#include <ncurses.h>
#include <wchar.h>

#include <stdio.h>

#include "zncurses.mdh"
#include "zncurses.pro"

static char *nam="zncurses";

#define ZNCURSES_MAX_WINDOWS 9

static WINDOW *zncurses_WIN[ZNCURSES_MAX_WINDOWS + 1];

/**/
static int
bin_zncurses_newwin(char *nam, char **args, Options ops, UNUSED(int func))
{
    WINDOW *w;
    int nlines, ncols, begin_y, begin_x;
    unsigned winnum;

    nlines = atoi(args[0]);
    ncols = atoi(args[1]);
    begin_y = atoi(args[2]);
    begin_x = atoi(args[3]);
    winnum = atoi(args[4]);

    if(winnum > ZNCURSES_MAX_WINDOWS) {
        zerrnam(nam, "bad window number: %s", args[4], 0);
        return 1;
    }

    if(zncurses_WIN[winnum]!=NULL) {
        zwarnnam(nam, "window number %s already defined", args[4], 0);
        return 1;
    }

    zncurses_WIN[winnum]=newwin(nlines, ncols, begin_y, begin_x);

    if(zncurses_WIN[winnum]==NULL)
	return 1;

    return 0;
}

/**/
static int
bin_zncurses_delwin(char *nam, char **args, Options ops, UNUSED(int func))
{
    unsigned winnum;

    winnum = atoi(args[0]);

    if(winnum > ZNCURSES_MAX_WINDOWS) {
        zerrnam(nam, "bad window number: %s", args[0], 0);
        return 1;
    }

    if(zncurses_WIN[winnum]==NULL) {
        zwarnnam(nam, "window number %s does not exist", args[0], 0);
        return 1;
    }

    if(delwin(zncurses_WIN[winnum])!=OK)
	return 1;

    zncurses_WIN[winnum]=NULL;
    return 0;
}

/**/
static int
bin_zncurses_wrefresh(char *nam, char **args, Options ops, UNUSED(int func))
{
    unsigned winnum;

    winnum = atoi(args[0]);

    if(winnum > ZNCURSES_MAX_WINDOWS) {
        zerrnam(nam, "bad window number: %s", args[0], 0);
        return 1;
    }

    if(zncurses_WIN[winnum]==NULL) {
        zwarnnam(nam, "window number %s does not exist", args[0], 0);
        return 1;
    }

    if(wrefresh(zncurses_WIN[winnum])!=OK)
	return 1;

    return 0;
}

/**/
static int
bin_zncurses_wmove(char *nam, char **args, Options ops, UNUSED(int func))
{
    WINDOW *w;
    int y, x;
    unsigned winnum;

    winnum = atoi(args[0]);
    y = atoi(args[1]);
    x = atoi(args[2]);

    if(winnum > ZNCURSES_MAX_WINDOWS) {
        zerrnam(nam, "bad window number: %s", args[0], 0);
        return 1;
    }

    if(zncurses_WIN[winnum]==NULL) {
        zwarnnam(nam, "window number %s is not defined", args[0], 0);
        return 1;
    }

    if(wmove(zncurses_WIN[winnum], y, x)!=OK)
	return 1;

    return 0;
}

/**/
static int
bin_zncurses_wadd_wch(char *nam, char **args, Options ops, UNUSED(int func))
{
    unsigned winnum;
    WINDOW *w;
    wchar_t c;
    cchar_t cc;

    winnum = atoi(args[0]);

    if(mbrtowc(&c, args[1], MB_CUR_MAX, NULL) < 1)
	return 1;

    if (setcchar(&cc, &c, A_NORMAL, 0, NULL)==ERR)
	return 1;

    if(winnum > ZNCURSES_MAX_WINDOWS) {
        zerrnam(nam, "bad window number: %s", args[0], 0);
        return 1;
    }

    if(zncurses_WIN[winnum]==NULL) {
        zwarnnam(nam, "window number %s is not defined", args[0], 0);
        return 1;
    }

    if(wadd_wch(zncurses_WIN[winnum], &cc)!=OK)
	return 1;

    return 0;
}

/**/
static int
bin_zncurses_wadd_wchstr(char *nam, char **args, Options ops, UNUSED(int func))
{
    unsigned winnum;
    WINDOW *w;
    wchar_t *ws;
    cchar_t *wcc;
    size_t sl;

    winnum = atoi(args[0]);

    if(winnum > ZNCURSES_MAX_WINDOWS) {
        zerrnam(nam, "bad window number: %s", args[0], 0);
        return 1;
    }

    if(zncurses_WIN[winnum]==NULL) {
        zwarnnam(nam, "window number %s is not defined", args[0], 0);
        return 1;
    }

    sl = strlen(args[1]);

    if(sl == 0) {
        return 0;
    }

    ws = malloc(sl * sizeof(wchar_t));

    if(mbstowcs(ws, args[1], sl) < 1) {
	free(ws);
	return 1;
    }

    wcc = malloc(wcslen(ws) * sizeof(cchar_t));

    if (setcchar(wcc, ws, A_NORMAL, 0, NULL)==ERR) {
	return 1;
    }

    free(ws);

    if(wadd_wchstr(zncurses_WIN[winnum], wcc)!=OK) {
        free(wcc);
	return 1;
    }

    free(wcc);
    return 0;
}

/**/
static int
bin_zncurses_wborder(char *nam, char **args, Options ops, UNUSED(int func))
{
    unsigned winnum;

    winnum = atoi(args[0]);

    if(winnum > ZNCURSES_MAX_WINDOWS) {
        zerrnam(nam, "bad window number: %s", args[0], 0);
        return 1;
    }

    if(zncurses_WIN[winnum]==NULL) {
        zwarnnam(nam, "window number %s does not exist", args[0], 0);
        return 1;
    }

    if(wborder(zncurses_WIN[winnum], 0, 0, 0, 0, 0, 0, 0, 0)!=OK)
	return 1;

    return 0;
}



/*
 * boot_ is executed when the module is loaded.
 */

static struct builtin bintab[] = {
    BUILTIN("zncurses_newwin", 0, bin_zncurses_newwin, 5, 5, 0, NULL, NULL),
    BUILTIN("zncurses_delwin", 0, bin_zncurses_delwin, 1, 1, 0, NULL, NULL),
    BUILTIN("zncurses_wrefresh", 0, bin_zncurses_wrefresh, 1, 1, 0, NULL, NULL),
    BUILTIN("zncurses_wmove", 0, bin_zncurses_wmove, 3, 3, 0, NULL, NULL),
    BUILTIN("zncurses_wadd_wch", 0, bin_zncurses_wadd_wch, 2, 2, 0, NULL, NULL),
    BUILTIN("zncurses_wadd_wchstr", 0, bin_zncurses_wadd_wchstr, 2, 2, 0, NULL, NULL),
    BUILTIN("zncurses_wborder", 0, bin_zncurses_wborder, 1, 1, 0, NULL, NULL),
};

/**/
int
setup_(UNUSED(Module m))
{
    return 0;
}

/**/
int
boot_(Module m)
{
    int i;
    for(i=1;i<=ZNCURSES_MAX_WINDOWS;i++)
	zncurses_WIN[i]=NULL;

    zncurses_WIN[0]=initscr();

    return !(addbuiltins(m->nam, bintab, sizeof(bintab)/sizeof(*bintab)));
}

/**/
int
cleanup_(Module m)
{
    deletebuiltins(m->nam, bintab, sizeof(bintab)/sizeof(*bintab));
    return 0;
}

/**/
int
finish_(UNUSED(Module m))
{
    return 0;
}
